@@ -5,16 +5,18 @@
 import sre_constants
 from PyQt4.QtGui import *
 from PyQt4.QtCore import *
+from PyQt4.QtWebKit import QWebPage
 import time, types, sys, re
-from operator import attrgetter
+from operator import attrgetter, itemgetter
 import anki, anki.utils, ankiqt.forms
 from ankiqt import ui
 from anki.cards import cardsTable, Card
 from anki.facts import factsTable, fieldsTable, Fact
 from anki.utils import fmtTimeSpan, parseTags, findTag, addTags, deleteTags, \
-     stripHTML, ids2str
+     stripHTMLAlt, ids2str
 from ankiqt.ui.utils import saveGeom, restoreGeom, saveSplitter, restoreSplitter
-from ankiqt.ui.utils import saveHeader, restoreHeader
+from ankiqt.ui.utils import saveHeader, restoreHeader, saveState, \
+     restoreState, applyStyles
 from anki.errors import *
 from anki.db import *
 from anki.stats import CardStats
@@ -30,7 +32,18 @@ CARD_CREATED = 6
 CARD_MODIFIED = 7
 CARD_INTERVAL = 8
 CARD_EASE = 9
-CARD_PRIORITY = 10
+CARD_NO = 10
+CARD_PRIORITY = 11
+CARD_TAGS = 12
+CARD_FACTCREATED = 13
+CARD_FIRSTANSWERED = 14
+
+COLOUR_SUSPENDED1 = "#ffffcc"
+COLOUR_SUSPENDED2 = "#ffffaa"
+COLOUR_INACTIVE1 = "#ffcccc"
+COLOUR_INACTIVE2 = "#ffaaaa"
+COLOUR_MARKED1 = "#ccccff"
+COLOUR_MARKED2 = "#aaaaff"
 
 # Deck editor
 ##########################################################################
@@ -69,21 +82,31 @@ class DeckModel(QAbstractTableModel):
             f = QFont()
             f.setPixelSize(self.parent.config['editFontSize'])
             return QVariant(f)
+        if role == Qt.TextAlignmentRole and index.column() == 2:
+            return QVariant(Qt.AlignHCenter)
         elif role == Qt.DisplayRole or role == Qt.EditRole:
             if len(self.cards[index.row()]) == 1:
                 # not cached yet
                 self.updateCard(index)
             s = self.columns[index.column()][1](index)
+            s = self.limitContent(s)
             s = s.replace("<br>", u" ")
             s = s.replace("<br />", u" ")
-            s = s.replace("\n", u"  ")
-            s = stripHTML(s)
+            s = s.replace("\n", u" ")
             s = re.sub("\[sound:[^]]+\]", "", s)
+            s = stripHTMLAlt(s)
             s = s.strip()
             return QVariant(s)
         else:
             return QVariant()
 
+    def limitContent(self, txt):
+        if "<c>" in txt:
+            matches = re.findall("(?s)<c>(.*?)</c>", txt)
+            return " ".join(matches)
+        else:
+            return txt
+
     def headerData(self, section, orientation, role):
         if orientation == Qt.Vertical:
             return QVariant()
@@ -167,15 +190,20 @@ where cards.factId = facts.id """
     def updateCard(self, index):
         try:
             self.cards[index.row()] = self.deck.s.first("""
-    select id, question, answer, due, reps, factId, created, modified,
-    interval, factor, priority from cards where id = :id""",
+select id, question, answer, combinedDue, reps, factId, created, modified,
+interval, factor, noCount, priority, (select tags from facts where
+facts.id = cards.factId), (select created from facts where
+facts.id = cards.factId), firstAnswered from cards where id = :id""",
                                                         id=self.cards[index.row()][0])
-            self.emit(SIGNAL("dataChanged(QModelIndex,QModelIndex)"),
-                      index, self.index(index.row(), 1))
+            self.emit(SIGNAL("layoutChanged()"))
         except:
             # called after search changed
             pass
 
+    def refresh(self):
+        self.cards = [[x[0]] for x in self.cards]
+        self.emit(SIGNAL("layoutChanged()"))
+
     # Tools
     ######################################################################
 
@@ -203,7 +231,7 @@ where cards.factId = facts.id """
         reps = self.cards[index.row()][CARD_REPS]
         secs = d - time.time()
         if secs <= 0:
-            if not reps:
+            if not reps and self.deck.newCardOrder == 0:
                 return _("(new card)")
             else:
                 return _("%s ago") % fmtTimeSpan(abs(secs), pad=0)
@@ -221,6 +249,12 @@ where cards.factId = facts.id """
             return self.repsColumn(index)
         elif self.sortKey == "factor":
             return self.easeColumn(index)
+        elif self.sortKey == "noCount":
+            return self.noColumn(index)
+        elif self.sortKey == "fact":
+            return self.factCreatedColumn(index)
+        elif self.sortKey == "firstAnswered":
+            return self.firstAnsweredColumn(index)
         else:
             return self.nextDue(index)
 
@@ -235,6 +269,12 @@ where cards.factId = facts.id """
             k = _("Reps")
         elif self.sortKey == "factor":
             k = _("Ease")
+        elif self.sortKey == "noCount":
+            k = _("Lapses")
+        elif self.sortKey == "firstAnswered":
+            k = _("First Answered")
+        elif self.sortKey == "fact":
+            k = _("Fact Created")
         else:
             k = _("Due")
         self.columns[-1][0] = k
@@ -243,6 +283,10 @@ where cards.factId = facts.id """
         return time.strftime("%Y-%m-%d", time.localtime(
             self.cards[index.row()][CARD_CREATED]))
 
+    def factCreatedColumn(self, index):
+        return time.strftime("%Y-%m-%d", time.localtime(
+            self.cards[index.row()][CARD_FACTCREATED]))
+
     def modifiedColumn(self, index):
         return time.strftime("%Y-%m-%d", time.localtime(
             self.cards[index.row()][CARD_MODIFIED]))
@@ -257,6 +301,17 @@ where cards.factId = facts.id """
     def easeColumn(self, index):
         return "%0.2f" % self.cards[index.row()][CARD_EASE]
 
+    def noColumn(self, index):
+        return "%d" % self.cards[index.row()][CARD_NO]
+
+    def firstAnsweredColumn(self, index):
+        firstAnswered = self.cards[index.row()][CARD_FIRSTANSWERED]
+        if firstAnswered == 0:
+            return _("(new card)")
+        else:
+            return time.strftime("%Y-%m-%d", time.localtime(firstAnswered))
+
+
 class StatusDelegate(QItemDelegate):
 
     def __init__(self, parent, model):
@@ -267,9 +322,20 @@ class StatusDelegate(QItemDelegate):
         if len(self.model.cards[index.row()]) == 1:
             self.model.updateCard(index)
         row = self.model.cards[index.row()]
-        if row[CARD_PRIORITY] == 0:
+        if row[CARD_PRIORITY] == -3:
             # custom render
-            brush = QBrush(QColor("#ffffaa"))
+            if index.row() % 2 == 0:
+                brush = QBrush(QColor(COLOUR_SUSPENDED1))
+            else:
+                brush = QBrush(QColor(COLOUR_SUSPENDED2))
+            painter.save()
+            painter.fillRect(option.rect, brush)
+            painter.restore()
+        elif "Marked" in row[CARD_TAGS]:
+            if index.row() % 2 == 0:
+                brush = QBrush(QColor(COLOUR_MARKED1))
+            else:
+                brush = QBrush(QColor(COLOUR_MARKED2))
             painter.save()
             painter.fillRect(option.rect, brush)
             painter.restore()
@@ -283,29 +349,37 @@ class EditDeck(QMainWindow):
         else:
             windParent = parent
         QMainWindow.__init__(self, windParent)
+        self.setAttribute(Qt.WA_Maemo5AutoOrientation, True)  #Screen rotation in N900
+        applyStyles(self)
         self.parent = parent
         self.deck = self.parent.deck
         self.config = parent.config
-        self.forceClose = False
         self.origModTime = parent.deck.modified
         self.currentRow = None
         self.lastFilter = ""
         self.dialog = ankiqt.forms.cardlist.Ui_MainWindow()
         self.dialog.setupUi(self)
-        restoreGeom(self, "editor")
+        self.setUnifiedTitleAndToolBarOnMac(True)
+        restoreGeom(self, "editor", 38)
+        restoreState(self, "editor")
         restoreSplitter(self.dialog.splitter, "editor")
+        self.dialog.splitter.setChildrenCollapsible(False)
+        # toolbar
+        self.dialog.toolBar.setIconSize(QSize(self.config['iconSize'],
+                                              self.config['iconSize']))
+        self.dialog.toolBar.toggleViewAction().setText(_("Toggle Toolbar"))
         # flush all changes before we load
         self.deck.s.flush()
         self.model = DeckModel(self.parent, self.parent.deck)
         self.dialog.tableView.setSortingEnabled(False)
+        self.dialog.tableView.setShowGrid(False)
         self.dialog.tableView.setModel(self.model)
         self.dialog.tableView.selectionModel()
         self.connect(self.dialog.tableView.selectionModel(),
                      SIGNAL("selectionChanged(QItemSelection,QItemSelection)"),
                      self.updateFilterLabel)
         self.dialog.tableView.setItemDelegate(StatusDelegate(self, self.model))
-        if self.deck.getInt("reverseOrder"):
-            self.dialog.actionReverseOrder.setChecked(True)
+        self.updateSortOrder()
         self.updateFont()
         self.setupMenus()
         self.setupFilter()
@@ -319,18 +393,13 @@ class EditDeck(QMainWindow):
         self.drawTags()
         self.updateFilterLabel()
         self.show()
-        self.updateSearch()
         if self.parent.currentCard:
             self.currentCard = self.parent.currentCard
-        self.focusCurrentCard()
-        if sys.platform.startswith("darwin"):
-            self.macCloseShortcut = QShortcut(QKeySequence("Ctrl+w"), self)
-            self.connect(self.macCloseShortcut, SIGNAL("activated()"),
-                         self.close)
+        self.updateSearch()
 
-    def findCardInDeckModel(self, model, card):
-        for i, thisCard in enumerate(model.cards):
-            if thisCard[0] == card.id:
+    def findCardInDeckModel(self):
+        for i, thisCard in enumerate(self.model.cards):
+            if thisCard[0] == self.currentCard.id:
                 return i
         return -1
 
@@ -361,21 +430,25 @@ class EditDeck(QMainWindow):
         self.connect(self.dialog.sortBox, SIGNAL("activated(int)"),
                      self.sortChanged)
         self.sortChanged(self.sortIndex, refresh=False)
+        self.connect(self.dialog.sortOrder, SIGNAL("clicked()"),
+                     self.reverseOrder)
 
     def drawTags(self):
-        self.dialog.tagList.view().setFixedWidth(200)
         self.dialog.tagList.setMaxVisibleItems(30)
-        self.dialog.tagList.setFixedWidth(130)
+        self.dialog.tagList.view().setMinimumWidth(200)
+        self.dialog.tagList.setFixedWidth(170)
         self.dialog.tagList.clear()
-        alltags = [None, "Marked", "Suspended", None, None, None]
+        alltags = [None, "Marked", None, None, "Leech", None, None]
         # system tags
-        self.dialog.tagList.addItem(_("<Filter>"))
+        self.dialog.tagList.addItem(_("Show All Cards"))
         self.dialog.tagList.addItem(QIcon(":/icons/rating.png"),
                                     _('Marked'))
         self.dialog.tagList.addItem(QIcon(":/icons/media-playback-pause.png"),
                                     _('Suspended'))
         self.dialog.tagList.addItem(QIcon(":/icons/chronometer.png"),
                                     _('Due'))
+        self.dialog.tagList.addItem(QIcon(":/icons/emblem-important.png"),
+                                    _('Leech'))
         self.dialog.tagList.addItem(QIcon(":/icons/editclear.png"),
                                     _('No fact tags'))
         self.dialog.tagList.insertSeparator(
@@ -423,6 +496,8 @@ class EditDeck(QMainWindow):
             _("Reps"),
             _("Ease"),
             _("Fact Created"),
+            _("Lapses"),
+            _("First Review"),
             ]
         self.sortFields = sorted(self.deck.allFields())
         self.sortList.extend([_("'%s'") % f for f in self.sortFields])
@@ -432,6 +507,12 @@ class EditDeck(QMainWindow):
             self.sortIndex = 0
         self.dialog.sortBox.setCurrentIndex(self.sortIndex)
 
+    def updateSortOrder(self):
+        if self.deck.getInt("reverseOrder"):
+            self.dialog.sortOrder.setIcon(QIcon(":/icons/view-sort-descending.png"))
+        else:
+            self.dialog.sortOrder.setIcon(QIcon(":/icons/view-sort-ascending.png"))
+
     def sortChanged(self, idx, refresh=True):
         if idx == 0:
             self.sortKey = "question"
@@ -442,7 +523,7 @@ class EditDeck(QMainWindow):
         elif idx == 3:
             self.sortKey = "modified"
         elif idx == 4:
-            self.sortKey = "due"
+            self.sortKey = "combinedDue"
         elif idx == 5:
             self.sortKey = "interval"
         elif idx == 6:
@@ -451,8 +532,12 @@ class EditDeck(QMainWindow):
             self.sortKey = "factor"
         elif idx == 8:
             self.sortKey = "fact"
+        elif idx == 9:
+            self.sortKey = "noCount"
+        elif idx == 10:
+            self.sortKey = "firstAnswered"
         else:
-            self.sortKey = ("field", self.sortFields[idx-9])
+            self.sortKey = ("field", self.sortFields[idx-11])
         self.rebuildSortIndex(self.sortKey)
         self.sortIndex = idx
         self.deck.setVar('sortIndex', idx)
@@ -467,7 +552,7 @@ class EditDeck(QMainWindow):
     def rebuildSortIndex(self, key):
         if key not in (
             "question", "answer", "created", "modified", "due", "interval",
-            "reps", "factor"):
+            "reps", "factor", "noCount", "firstAnswered"):
             return
         old = self.deck.s.scalar("select sql from sqlite_master where name = :k",
                                  k="ix_cards_sort")
@@ -492,10 +577,12 @@ class EditDeck(QMainWindow):
         elif idx == 1:
             filter = "tag:marked"
         elif idx == 2:
-            filter = "tag:suspended"
+            filter = "is:suspended"
         elif idx == 3:
             filter = "is:due"
         elif idx == 4:
+            filter = "tag:leech"
+        elif idx == 5:
             filter = "tag:none"
         else:
             filter = "tag:" + self.alltags[idx]
@@ -512,7 +599,7 @@ class EditDeck(QMainWindow):
             "cur": len(self.model.cards),
             "tot": self.deck.cardCount,
             "sel": ngettext("%d selected", "%d selected", selected) % selected
-            } + " - " + self.parent.deck.name())
+            } + " - " + self.deck.name())
 
     def onEvent(self, type='field'):
         if self.deck.undoAvailable():
@@ -527,11 +614,14 @@ class EditDeck(QMainWindow):
             self.dialog.actionRedo.setEnabled(True)
         else:
             self.dialog.actionRedo.setEnabled(False)
-        # update list
-        if self.currentRow and self.model.cards:
-            self.model.updateCard(self.currentRow)
-        if type == "tag":
-            self.drawTags()
+        if type=="all":
+            self.updateAfterCardChange()
+        else:
+            # update list
+            if self.currentRow and self.model.cards:
+                self.model.updateCard(self.currentRow)
+            if type == "tag":
+                self.drawTags()
 
     def filterTextChanged(self):
         interval = 300
@@ -556,30 +646,30 @@ class EditDeck(QMainWindow):
     def updateSearch(self, force=True):
         if self.parent.inDbHandler:
             return
-        idx = self.dialog.tableView.currentIndex()
-        row = idx.row()
         self.model.searchStr = unicode(self.dialog.filterEdit.text())
         self.model.showMatching(force)
         self.updateFilterLabel()
         self.onEvent()
         self.filterTimer = None
         if self.model.cards:
-            if row == -1:
-                row = 0
             self.dialog.cardInfoGroup.show()
             self.dialog.fieldsArea.show()
         else:
             self.dialog.cardInfoGroup.hide()
             self.dialog.fieldsArea.hide()
-        self.dialog.tableView.selectRow(row)
-        self.dialog.tableView.scrollTo(idx, QAbstractItemView.PositionAtCenter)
+        if not self.focusCurrentCard():
+            if self.model.cards:
+                self.dialog.tableView.selectRow(0)
         if not self.model.cards:
             self.editor.setFact(None)
 
     def focusCurrentCard(self):
         if self.currentCard:
-            currentCardIndex = self.findCardInDeckModel(
-                                 self.model, self.currentCard)
+            try:
+                self.currentCard.id
+            except:
+                return False
+            currentCardIndex = self.findCardInDeckModel()
             if currentCardIndex >= 0:
                 sm = self.dialog.tableView.selectionModel()
                 sm.clear()
@@ -587,6 +677,8 @@ class EditDeck(QMainWindow):
                 self.dialog.tableView.scrollTo(
                               self.model.index(currentCardIndex,0),
                               self.dialog.tableView.PositionAtCenter)
+                return True
+        return False
 
     def setupHeaders(self):
         if not sys.platform.startswith("win32"):
@@ -599,6 +691,7 @@ class EditDeck(QMainWindow):
 
     def setupMenus(self):
         # actions
+        self.connect(self.dialog.actionAddItems, SIGNAL("triggered()"), self.parent.onAddCard)
         self.connect(self.dialog.actionDelete, SIGNAL("triggered()"), self.deleteCards)
         self.connect(self.dialog.actionAddTag, SIGNAL("triggered()"), self.addTags)
         self.connect(self.dialog.actionDeleteTag, SIGNAL("triggered()"), self.deleteTags)
@@ -606,16 +699,16 @@ class EditDeck(QMainWindow):
         self.connect(self.dialog.actionCram, SIGNAL("triggered()"), self.cram)
         self.connect(self.dialog.actionAddCards, SIGNAL("triggered()"), self.addCards)
         self.connect(self.dialog.actionChangeModel, SIGNAL("triggered()"), self.onChangeModel)
-        self.connect(self.dialog.actionSuspend, SIGNAL("triggered()"), self.onSuspend)
-        self.connect(self.dialog.actionUnsuspend, SIGNAL("triggered()"), self.onUnsuspend)
+        self.connect(self.dialog.actionToggleSuspend, SIGNAL("triggered(bool)"), self.onSuspend)
+        self.connect(self.dialog.actionToggleMark, SIGNAL("triggered(bool)"), self.onMark)
         # edit
         self.connect(self.dialog.actionFont, SIGNAL("triggered()"), self.onFont)
         self.connect(self.dialog.actionUndo, SIGNAL("triggered()"), self.onUndo)
         self.connect(self.dialog.actionRedo, SIGNAL("triggered()"), self.onRedo)
         self.connect(self.dialog.actionInvertSelection, SIGNAL("triggered()"), self.invertSelection)
-        self.connect(self.dialog.actionReverseOrder, SIGNAL("triggered()"), self.reverseOrder)
         self.connect(self.dialog.actionSelectFacts, SIGNAL("triggered()"), self.selectFacts)
         self.connect(self.dialog.actionFindReplace, SIGNAL("triggered()"), self.onFindReplace)
+        self.connect(self.dialog.actionFindDuplicates, SIGNAL("triggered()"), self.onFindDupes)
         # jumps
         self.connect(self.dialog.actionFirstCard, SIGNAL("triggered()"), self.onFirstCard)
         self.connect(self.dialog.actionLastCard, SIGNAL("triggered()"), self.onLastCard)
@@ -625,6 +718,7 @@ class EditDeck(QMainWindow):
         self.connect(self.dialog.actionFact, SIGNAL("triggered()"), self.onFact)
         self.connect(self.dialog.actionTags, SIGNAL("triggered()"), self.onTags)
         self.connect(self.dialog.actionSort, SIGNAL("triggered()"), self.onSort)
+        self.connect(self.dialog.actionCardList, SIGNAL("triggered()"), self.onCardList)
         # help
         self.connect(self.dialog.actionGuide, SIGNAL("triggered()"), self.onHelp)
         runHook('editor.setupMenus', self)
@@ -632,19 +726,13 @@ class EditDeck(QMainWindow):
     def onClose(self):
         saveSplitter(self.dialog.splitter, "editor")
         self.editor.saveFieldsNow()
-        if not self.forceClose:
-            if not self.factValid:
-                ui.utils.showInfo(_(
-                    "Some fields are missing or not unique."),
-                                  parent=self, help="AddItems#AddError")
-                return
         self.editor.setFact(None)
         self.editor.close()
         saveGeom(self, "editor")
+        saveState(self, "editor")
         saveHeader(self.dialog.tableView.horizontalHeader(), "editor")
         self.hide()
         ui.dialogs.close("CardList")
-        self.parent.moveToState("auto")
         self.teardownHooks()
         return True
 
@@ -666,32 +754,11 @@ class EditDeck(QMainWindow):
         self.editor = ui.facteditor.FactEditor(self,
                                                self.dialog.fieldsArea,
                                                self.deck)
-        self.factValid = True
-        self.editor.onFactValid = self.onFactValid
-        self.editor.onFactInvalid = self.onFactInvalid
         self.editor.onChange = self.onEvent
         self.connect(self.dialog.tableView.selectionModel(),
                      SIGNAL("currentRowChanged(QModelIndex, QModelIndex)"),
                      self.rowChanged)
 
-    def onFactValid(self, fact):
-        self.factValid = True
-        self.dialog.tableView.setEnabled(True)
-        self.dialog.filterEdit.setEnabled(True)
-        self.dialog.sortBox.setEnabled(True)
-        self.dialog.tagList.setEnabled(True)
-        self.dialog.menubar.setEnabled(True)
-        self.dialog.cardInfoGroup.setEnabled(True)
-
-    def onFactInvalid(self, fact):
-        self.factValid = False
-        self.dialog.tableView.setEnabled(False)
-        self.dialog.filterEdit.setEnabled(False)
-        self.dialog.sortBox.setEnabled(False)
-        self.dialog.tagList.setEnabled(False)
-        self.dialog.menubar.setEnabled(False)
-        self.dialog.cardInfoGroup.setEnabled(False)
-
     def rowChanged(self, current, previous):
         self.currentRow = current
         self.currentCard = self.model.getCard(current)
@@ -700,8 +767,10 @@ class EditDeck(QMainWindow):
             return
         fact = self.currentCard.fact
         self.editor.setFact(fact, True)
+        self.editor.card = self.currentCard
         self.showCardInfo(self.currentCard)
         self.onEvent()
+        self.updateToggles()
 
     def setupCardInfo(self):
         self.currentCard = None
@@ -731,14 +800,13 @@ where id in (%s)""" % ",".join([
             "select id from cards where factId in (%s)" %
             ",".join([str(s) for s in self.selectedFacts()]))
 
-    def updateAfterCardChange(self, reset=False):
-        "Refresh info like stats on current card"
+    def updateAfterCardChange(self):
+        "Refresh info like stats on current card, and rebuild mw queue."
         self.currentRow = self.dialog.tableView.currentIndex()
         self.rowChanged(self.currentRow, None)
-        if reset:
-            self.updateSearch()
+        self.model.refresh()
         self.drawTags()
-        self.parent.moveToState("auto")
+        self.parent.reset()
 
     # Menu options
     ######################################################################
@@ -746,51 +814,108 @@ where id in (%s)""" % ",".join([
     def deleteCards(self):
         cards = self.selectedCards()
         n = _("Delete Cards")
+        try:
+            new = self.findCardInDeckModel() + 1
+        except:
+            # card has been deleted
+            return
+        # ensure the change timer doesn't fire after deletion but before reset
+        self.editor.saveFieldsNow()
+        self.editor.fact = None
+        self.dialog.tableView.setFocus()
         self.deck.setUndoStart(n)
         self.deck.deleteCards(cards)
         self.deck.setUndoEnd(n)
+        new = min(max(0, new), len(self.model.cards) - 1)
+        self.dialog.tableView.selectRow(new)
         self.updateSearch()
         self.updateAfterCardChange()
 
-    def addTags(self):
-        (tags, r) = ui.utils.getTag(self, self.deck, _("Enter tags to add:"))
-        if tags:
-            n = _("Add Tags")
+    def addTags(self, tags=None, label=None):
+        # focus lost hook may not have chance to fire
+        self.editor.saveFieldsNow()
+        if tags is None:
+            (tags, r) = ui.utils.getTag(self, self.deck, _("Enter tags to add:"))
+        else:
+            r = True
+        if label is None:
+            label = _("Add Tags")
+        if r:
             self.parent.setProgressParent(self)
-            self.deck.setUndoStart(n)
+            self.deck.setUndoStart(label)
             self.deck.addTags(self.selectedFacts(), tags)
-            self.deck.setUndoEnd(n)
+            self.deck.setUndoEnd(label)
             self.parent.setProgressParent(None)
-        self.updateAfterCardChange(reset=True)
+        self.updateAfterCardChange()
 
-    def deleteTags(self):
-        (tags, r) = ui.utils.getTag(self, self.deck, _("Enter tags to delete:"))
-        if tags:
-            n = _("Delete Tags")
+    def deleteTags(self, tags=None, label=None):
+        # focus lost hook may not have chance to fire
+        self.editor.saveFieldsNow()
+        if tags is None:
+            (tags, r) = ui.utils.getTag(self, self.deck, _("Enter tags to delete:"))
+        else:
+            r = True
+        if label is None:
+            label = _("Delete Tags")
+        if r:
             self.parent.setProgressParent(self)
-            self.deck.setUndoStart(n)
+            self.deck.setUndoStart(label)
             self.deck.deleteTags(self.selectedFacts(), tags)
-            self.deck.setUndoEnd(n)
+            self.deck.setUndoEnd(label)
             self.parent.setProgressParent(None)
-        self.updateAfterCardChange(reset=True)
+        self.updateAfterCardChange()
 
-    def onSuspend(self):
+    def updateToggles(self):
+        self.dialog.actionToggleSuspend.setChecked(self.isSuspended())
+        self.dialog.actionToggleMark.setChecked(self.isMarked())
+
+    def isSuspended(self):
+        return self.currentCard and self.currentCard.type < 0
+
+    def onSuspend(self, sus):
+        # focus lost hook may not have chance to fire
+        self.editor.saveFieldsNow()
+        if sus:
+            self._onSuspend()
+        else:
+            self._onUnsuspend()
+
+    def _onSuspend(self):
         n = _("Suspend")
         self.parent.setProgressParent(self)
         self.deck.setUndoStart(n)
-        self.deck.addTags(self.selectedFacts(), "Suspended")
+        self.deck.suspendCards(self.selectedCards())
+        self.parent.reset()
         self.deck.setUndoEnd(n)
         self.parent.setProgressParent(None)
-        self.updateAfterCardChange(reset=True)
+        self.model.refresh()
+        self.updateAfterCardChange()
 
-    def onUnsuspend(self):
+    def _onUnsuspend(self):
         n = _("Unsuspend")
         self.parent.setProgressParent(self)
         self.deck.setUndoStart(n)
-        self.deck.deleteTags(self.selectedFacts(), "Suspended")
+        self.deck.unsuspendCards(self.selectedCards())
+        self.parent.reset()
         self.deck.setUndoEnd(n)
         self.parent.setProgressParent(None)
-        self.updateAfterCardChange(reset=True)
+        self.model.refresh()
+        self.updateAfterCardChange()
+
+    def isMarked(self):
+        return self.currentCard and "Marked" in self.currentCard.fact.tags
+
+    def onMark(self, mark):
+        if mark:
+            self._onMark()
+        else:
+            self._onUnmark()
+
+    def _onMark(self):
+        self.addTags(tags="Marked", label=_("Toggle Mark"))
+
+    def _onUnmark(self):
+        self.deleteTags(tags="Marked", label=_("Toggle Mark"))
 
     def reschedule(self):
         n = _("Reschedule")
@@ -805,19 +930,18 @@ where id in (%s)""" % ",".join([
                 self.deck.resetCards(self.selectedCards())
             else:
                 try:
-                    min = float(str(frm.rangeMin.text()))
-                    max = float(str(frm.rangeMax.text()))
+                    min = float(frm.rangeMin.value())
+                    max = float(frm.rangeMax.value())
                 except ValueError:
                     ui.utils.showInfo(
-                        _("Please enter a valid start and end range."),
+                        _("Please enter a valid range."),
                         parent=self)
                     return
                 self.deck.rescheduleCards(self.selectedCards(), min, max)
         finally:
-            self.deck.rebuildCounts(full=False)
-            self.deck.rebuildQueue()
+            self.deck.reset()
             self.deck.setUndoEnd(n)
-        self.updateAfterCardChange(reset=True)
+        self.updateAfterCardChange()
 
     def addCards(self):
         sf = self.selectedFacts()
@@ -845,12 +969,13 @@ where id in %s""" % ids2str(sf))
         facts = self.deck.s.query(Fact).filter(
             text("id in %s" % ids2str(sf))).order_by(Fact.created).all()
         self.deck.updateProgress(_("Generating Cards..."))
+        ids = []
         for c, fact in enumerate(facts):
-            self.deck.addCards(fact, d.selectedCms)
+            ids.extend(self.deck.addCards(fact, d.selectedCms))
             if c % 50 == 0:
                 self.deck.updateProgress()
         self.deck.flushMod()
-        self.deck.updateAllPriorities()
+        self.deck.updatePriorities(ids)
         self.deck.finishProgress()
         self.parent.setProgressParent(None)
         self.deck.setUndoEnd(n)
@@ -858,12 +983,8 @@ where id in %s""" % ids2str(sf))
         self.updateAfterCardChange()
 
     def cram(self):
-        if ui.utils.askUser(
-            _("Cram selected cards in new deck?"),
-            help="CramMode",
-            parent=self):
-            self.close()
-            self.parent.onCram(self.selectedCards())
+        self.close()
+        self.parent.onCram(self.selectedCards())
 
     def onChangeModel(self):
         sf = self.selectedFacts()
@@ -892,13 +1013,20 @@ where id in %s""" % ids2str(sf))
     ######################################################################
 
     def selectFacts(self):
+        self.deck.startProgress()
         sm = self.dialog.tableView.selectionModel()
+        sm.blockSignals(True)
         cardIds = dict([(x, 1) for x in self.selectedFactsAsCards()])
         for i, card in enumerate(self.model.cards):
-            if card.id in cardIds:
+            if card[0] in cardIds:
                 sm.select(self.model.index(i, 0),
                           QItemSelectionModel.Select | QItemSelectionModel.Rows)
-
+            if i % 100 == 0:
+                self.deck.updateProgress()
+        sm.blockSignals(False)
+        self.deck.finishProgress()
+        self.updateFilterLabel()
+        self.updateAfterCardChange()
 
     def invertSelection(self):
         sm = self.dialog.tableView.selectionModel()
@@ -911,6 +1039,7 @@ where id in %s""" % ids2str(sf))
         self.model.cards.reverse()
         self.model.reset()
         self.focusCurrentCard()
+        self.updateSortOrder()
 
     # Edit: undo/redo
     ######################################################################
@@ -1014,6 +1143,81 @@ where id in %s""" % ids2str(sf))
     def onFindReplaceHelp(self):
         QDesktopServices.openUrl(QUrl(ankiqt.appWiki +
                                       "Browser#FindReplace"))
+    # Edit: finding dupes
+    ######################################################################
+
+    def onFindDupes(self):
+        win = QDialog(self)
+        dialog = ankiqt.forms.finddupes.Ui_Dialog()
+        dialog.setupUi(win)
+        restoreGeom(win, "findDupes")
+        fields = sorted(self.currentCard.fact.model.fieldModels, key=attrgetter("name"))
+        # per-model data
+        data = self.deck.s.all("""
+select fm.id, m.name || '>' || fm.name from fieldmodels fm, models m
+where fm.modelId = m.id""")
+        data.sort(key=itemgetter(1))
+        # all-model data
+        data2 = self.deck.s.all("""
+select fm.id, fm.name from fieldmodels fm""")
+        byName = {}
+        for d in data2:
+            if d[1] in byName:
+                byName[d[1]].append(d[0])
+            else:
+                byName[d[1]] = [d[0]]
+        names = byName.keys()
+        names.sort()
+        alldata = [(byName[n], n) for n in names] + data
+        dialog.searchArea.addItems(QStringList([d[1] for d in alldata]))
+        # links
+        dialog.webView.page().setLinkDelegationPolicy(
+            QWebPage.DelegateAllLinks)
+        self.connect(dialog.webView,
+                     SIGNAL("linkClicked(QUrl)"),
+                     self.dupeLinkClicked)
+
+        def onFin(code):
+            saveGeom(win, "findDupes")
+        self.connect(win, SIGNAL("finished(int)"), onFin)
+
+        def onClick():
+            idx = dialog.searchArea.currentIndex()
+            data = alldata[idx]
+            if isinstance(data[0], list):
+                # all models
+                fmids = data[0]
+            else:
+                # single model
+                fmids = [data[0]]
+            self.duplicatesReport(dialog.webView, fmids)
+
+        self.connect(dialog.searchButton, SIGNAL("clicked()"),
+                     onClick)
+        win.show()
+
+    def duplicatesReport(self, web, fmids):
+        self.deck.startProgress(2)
+        self.deck.updateProgress(_("Finding..."))
+        res = self.deck.findDuplicates(fmids)
+        t = "<html><body>"
+        t += _("Duplicate Groups: %d") % len(res)
+        t += "<p><ol>"
+
+        for group in res:
+            t += '<li><a href="%s">%s</a>' % (
+                "fid:" + ",".join(str(id) for id in group[1]),
+                group[0])
+
+        t += "</ol>"
+        t += "</body></html>"
+        web.setHtml(t)
+        self.deck.finishProgress()
+
+    def dupeLinkClicked(self, link):
+        self.dialog.filterEdit.setText(str(link.toString()))
+        self.updateSearch()
+        self.onFact()
 
     # Jumping
     ######################################################################
@@ -1052,6 +1256,7 @@ where id in %s""" % ids2str(sf))
 
     def onFind(self):
         self.dialog.filterEdit.setFocus()
+        self.dialog.filterEdit.selectAll()
 
     def onFact(self):
         self.editor.focusFirst()
@@ -1062,6 +1267,9 @@ where id in %s""" % ids2str(sf))
     def onSort(self):
         self.dialog.sortBox.setFocus()
 
+    def onCardList(self):
+        self.dialog.tableView.setFocus()
+
     # Help
     ######################################################################
 
@@ -1075,6 +1283,7 @@ class AddCardChooser(QDialog):
 
     def __init__(self, parent, cms):
         QDialog.__init__(self, parent, Qt.Window)
+        self.setAttribute(Qt.WA_Maemo5AutoOrientation, True)  #Screen rotation in N900
         self.parent = parent
         self.cms = cms
         self.dialog = ankiqt.forms.addcardmodels.Ui_Dialog()
@@ -1121,7 +1330,9 @@ class ChangeModelDialog(QDialog):
 
     def __init__(self, parent, oldModel, oldTemplate):
         QDialog.__init__(self, parent, Qt.Window)
+        self.setAttribute(Qt.WA_Maemo5AutoOrientation, True)  #Screen rotation in N900
         self.parent = parent
+        self.origModel = self.parent.deck.currentModel
         self.oldModel = oldModel
         self.oldTemplate = oldTemplate
         self.form = ankiqt.forms.changemodel.Ui_Dialog()
@@ -1241,15 +1452,29 @@ class ChangeModelDialog(QDialog):
             combos=self.fieldCombos,
             new=self.targetModel.fieldModels)
 
+    def reject(self):
+        self.parent.deck.currentModel = self.origModel
+        self.modelChooser.deinit()
+        return QDialog.reject(self)
+
     def accept(self):
         saveGeom(self, "changeModel")
+        self.parent.deck.currentModel = self.origModel
         # check maps
         fmap = self.getFieldMap()
         cmap = self.getTemplateMap()
         if not cmap or (self.targetModel != self.oldModel and
                         not fmap):
-            return ui.utils.showInfo(
+            ui.utils.showInfo(
                 _("Targets must be unique."), parent=self)
+            return
+        if [c for c in cmap.values() if not c]:
+            if not ui.utils.askUser(_("""\
+Any cards with templates mapped to nothing will be deleted.
+If a fact has no remaining cards, it will be lost.
+Are you sure you want to continue?"""), parent=self):
+                return
+        self.modelChooser.deinit()
         if self.targetModel == self.oldModel:
             self.ret = (self.targetModel, None, cmap)
             return QDialog.accept(self)
