@@ -4,14 +4,25 @@
 from PyQt4.QtGui import *
 from PyQt4.QtCore import *
 import anki, ankiqt
-from anki.exporting import exporters
+from anki.exporting import exporters as exporters_
 from anki.utils import parseTags
 from ankiqt import ui
 
+class PackagedAnkiExporter(object):
+    def __init__(self, *args):
+        pass
+
+def exporters():
+    l = list(exporters_())
+    l.insert(1, (_("Packaged Anki Deck (*.zip)"),
+                 PackagedAnkiExporter))
+    return l
+
 class ExportDialog(QDialog):
 
     def __init__(self, parent):
         QDialog.__init__(self, parent, Qt.Window)
+        self.setAttribute(Qt.WA_Maemo5AutoOrientation, True)  #Screen rotation in N900
         self.parent = parent
         self.deck = parent.deck
         self.dialog = ankiqt.forms.exporting.Ui_ExportDialog()
@@ -35,7 +46,7 @@ class ExportDialog(QDialog):
         self.setTabOrder(self.tags,
                          self.dialog.includeScheduling)
         # save button
-        b = QPushButton(_("Export to..."))
+        b = QPushButton(_("Export..."))
         self.dialog.buttonBox.addButton(b, QDialogButtonBox.AcceptRole)
 
     def exporterChanged(self, idx):
@@ -50,6 +61,9 @@ class ExportDialog(QDialog):
             self.dialog.includeTags.hide()
 
     def accept(self):
+        if isinstance(self.exporter, PackagedAnkiExporter):
+            self.parent.onShare(unicode(self.tags.text()))
+            return QDialog.accept(self)
         file = ui.utils.getSaveFile(self, _("Choose file to export to"), "export",
                                     self.exporter.key, self.exporter.ext)
         self.hide()
