@@ -1,11 +1,12 @@
 # Copyright: Damien Elmes <anki@ichi2.net>
 # License: GNU GPL, version 3 or later; http://www.gnu.org/copyleft/gpl.html
 
-import os, copy, time
+import os, copy, time, sys, re, traceback
 from PyQt4.QtGui import *
 from PyQt4.QtCore import *
 import anki
 import anki.importing as importing
+from ankiqt.ui.utils import getOnlyText
 from anki.errors import *
 import ankiqt.forms
 from ankiqt import ui
@@ -13,21 +14,27 @@ from ankiqt import ui
 class ChangeMap(QDialog):
     def __init__(self, parent, model, current):
         QDialog.__init__(self, parent, Qt.Window)
+        self.setAttribute(Qt.WA_Maemo5AutoOrientation, True)  #Screen rotation in N900
         self.parent = parent
         self.model = model
         self.dialog = ankiqt.forms.changemap.Ui_ChangeMap()
         self.dialog.setupUi(self)
         n = 0
+        setCurrent = False
         for field in self.model.fieldModels:
             item = QListWidgetItem(_("Map to %s") % field.name)
             self.dialog.fields.addItem(item)
-            if current == field.name:
+            if current and current.name == field.name:
+                setCurrent = True
                 self.dialog.fields.setCurrentRow(n)
             n += 1
         self.dialog.fields.addItem(QListWidgetItem(_("Map to Tags")))
         self.dialog.fields.addItem(QListWidgetItem(_("Discard field")))
-        if current is None:
-            self.dialog.fields.setCurrentRow(n)
+        if not setCurrent:
+            if current == 0:
+                self.dialog.fields.setCurrentRow(n)
+            else:
+                self.dialog.fields.setCurrentRow(n+1)
         self.field = None
 
     def getField(self):
@@ -44,6 +51,30 @@ class ChangeMap(QDialog):
             self.field = 0
         QDialog.accept(self)
 
+class UpdateMap(QDialog):
+    def __init__(self, parent, numFields, fieldModels):
+        QDialog.__init__(self, parent, Qt.Window)
+        self.parent = parent
+        self.fieldModels = fieldModels
+        self.dialog = ankiqt.forms.importup.Ui_Dialog()
+        self.dialog.setupUi(self)
+        self.connect(self.dialog.buttonBox.button(QDialogButtonBox.Help),
+                     SIGNAL("clicked()"), self.helpRequested)
+        for i in range(numFields):
+            self.dialog.fileField.addItem("Field %d" % (i+1))
+        for m in fieldModels:
+            self.dialog.deckField.addItem(m.name)
+        self.exec_()
+
+    def helpRequested(self):
+        QDesktopServices.openUrl(QUrl(ankiqt.appWiki + "FileImport"))
+
+    def accept(self):
+        self.updateKey = (
+            self.dialog.fileField.currentIndex(),
+            self.fieldModels[self.dialog.deckField.currentIndex()].id)
+        QDialog.accept(self)
+
 class ImportDialog(QDialog):
 
     def __init__(self, parent):
@@ -51,51 +82,56 @@ class ImportDialog(QDialog):
         self.parent = parent
         self.dialog = ankiqt.forms.importing.Ui_ImportDialog()
         self.dialog.setupUi(self)
-        self.tags = ui.tagedit.TagEdit(parent)
-        self.tags.setDeck(parent.deck)
-        self.dialog.topGrid.addWidget(self.tags,2,1,1,1)
+        self.connect(self.dialog.buttonBox.button(QDialogButtonBox.Help),
+                     SIGNAL("clicked()"), self.helpRequested)
         self.setupMappingFrame()
         self.setupOptions()
+        self.getFile()
+        if not self.file:
+            return
+        self.dialog.groupBox.setTitle(os.path.basename(self.file))
+        self.maybePreview()
+        self.connect(self.dialog.autoDetect, SIGNAL("clicked()"),
+                     self.onDelimiter)
+        self.updateDelimiterButtonText()
         self.exec_()
 
     def setupOptions(self):
-        self.file = None
         self.model = self.parent.deck.currentModel
         self.modelChooser = ui.modelchooser.ModelChooser(self,
                                                          self.parent,
                                                          self.parent.deck,
                                                          self.modelChanged)
-        self.importerChanged(0)
-        self.connect(self.dialog.type, SIGNAL("activated(int)"),
-                     self.importerChanged)
-        self.dialog.type.insertItems(0, QStringList(list(zip(*importing.Importers)[0])))
-        self.connect(self.dialog.file, SIGNAL("clicked()"),
-                     self.changeFile)
         self.dialog.modelArea.setLayout(self.modelChooser)
         self.connect(self.dialog.importButton, SIGNAL("clicked()"),
                      self.doImport)
-        self.maybePreview()
+        self.connect(self.dialog.updateButton, SIGNAL("clicked()"),
+                     self.doUpdate)
 
-    def importerChanged(self, idx):
-        self.importerFunc = zip(*importing.Importers)[1][idx]
-        if self.importerFunc.needMapper:
-            self.modelChooser.show()
-            self.dialog.tagDuplicates.show()
-        else:
-            self.modelChooser.hide()
-            self.dialog.tagDuplicates.hide()
-        self.dialog.file.setText(_("Choose file..."))
-        self.file = None
-        self.maybePreview()
-
-    def changeFile(self):
-        key = zip(*importing.Importers)[0][self.dialog.type.currentIndex()]
-        file = ui.utils.getFile(self, _("Import file"), "import", key)
+    def getFile(self):
+        key = ";;".join([x[0] for x in importing.Importers])
+        file = ui.utils.getFile(self.parent, _("Import"), "import", key)
         if not file:
+            self.file = None
             return
         self.file = unicode(file)
-        self.dialog.file.setText(os.path.basename(self.file))
-        self.maybePreview()
+        ext = os.path.splitext(self.file)[1]
+        self.importer = None
+        for i in importing.Importers:
+            for mext in re.findall("[( ]?\*\.(.+?)[) ]", i[0]):
+                if ext == "." + mext:
+                    self.importer = i
+                    break
+        if not self.importer:
+            self.importer = importing.Importers[0]
+        self.importerFunc = self.importer[1]
+        if self.importerFunc.needMapper:
+            self.modelChooser.show()
+        else:
+            self.modelChooser.hide()
+            self.dialog.groupBox.setShown(False)
+            self.dialog.mappingGroup.setTitle("")
+        self.dialog.autoDetect.setShown(self.importerFunc.needDelimiter)
 
     def maybePreview(self):
         if self.file and self.model:
@@ -108,12 +144,59 @@ class ImportDialog(QDialog):
         self.model = model
         self.maybePreview()
 
-    def doImport(self):
+    def onDelimiter(self):
+        str = getOnlyText(_("""\
+By default, Anki will detect the character between fields, such as
+a tab, comma, and so on. If Anki is detecting the character incorrectly,
+you can enter it here. Use \\t to represent tab."""),
+                self, help="FileImport")
+        str = str.replace("\\t", "\t")
+        str = str.encode("ascii")
+        self.hideMapping()
+        def updateDelim():
+            self.importer.delimiter = str
+        self.showMapping(hook=updateDelim)
+        self.updateDelimiterButtonText()
+
+    def updateDelimiterButtonText(self):
+        if not self.importerFunc.needDelimiter:
+            return
+        if self.importer.delimiter:
+            d = self.importer.delimiter
+        else:
+            d = self.importer.dialect.delimiter
+        if d == "\t":
+            d = "Tab"
+        elif d == ",":
+            d = "Comma"
+        elif d == " ":
+            d = "Space"
+        elif d == ";":
+            d = "Semicolon"
+        elif d == ":":
+            d = "Colon"
+        else:
+            d = `d`
+        if self.importer.delimiter:
+            txt = _("Manual &delimiter: %s") % d
+        else:
+            txt = _("Auto-detected &delimiter: %s") % d
+        self.dialog.autoDetect.setText(txt)
+
+    def doUpdate(self):
+        f = UpdateMap(self,
+                      self.importer.fields(),
+                      self.model.fieldModels)
+        if not getattr(f, "updateKey", None):
+            # user cancelled
+            return
+        self.importer.updateKey = f.updateKey
+        self.doImport(True)
+
+    def doImport(self, update=False):
         self.dialog.status.setText(_("Importing..."))
         t = time.time()
         self.importer.mapping = self.mapping
-        self.importer.tagsToAdd = unicode(self.tags.text())
-        self.importer.tagDuplicates = self.dialog.tagDuplicates.isChecked()
         try:
             n = _("Import")
             self.parent.deck.setUndoStart(n)
@@ -125,36 +208,45 @@ class ImportDialog(QDialog):
                 self.dialog.status.setText(msg)
                 return
             except Exception, e:
-                msg = _("Import failed: %s\n") % `e`
+                msg = _("Import failed.\n")
+                msg += unicode(traceback.format_exc(), "ascii", "replace")
                 self.dialog.status.setText(msg)
                 return
         finally:
             self.parent.deck.finishProgress()
             self.parent.deck.setUndoEnd(n)
         txt = (
-            _("Importing complete. %(num)d cards imported from %(file)s.\n") %
+            _("Importing complete. %(num)d facts imported from %(file)s.\n") %
             {"num": self.importer.total, "file": os.path.basename(self.file)})
-        txt += _("Click the close button or import another file.\n\n")
+        self.dialog.groupBox.setShown(False)
+        self.dialog.buttonBox.button(QDialogButtonBox.Close).setFocus()
         if self.importer.log:
             txt += _("Log of import:\n") + "\n".join(self.importer.log)
         self.dialog.status.setText(txt)
         self.file = None
         self.maybePreview()
+        self.parent.deck.s.flush()
         self.parent.reset()
+        self.modelChooser.deinit()
 
     def setupMappingFrame(self):
         # qt seems to have a bug with adding/removing from a grid, so we add
         # to a separate object and add/remove that instead
-        self.mapbox = QVBoxLayout(self.dialog.mappingArea)
+        self.frame = QFrame(self.dialog.mappingArea)
+        self.dialog.mappingArea.setWidget(self.frame)
+        self.mapbox = QVBoxLayout(self.frame)
+        self.mapbox.setContentsMargins(0,0,0,0)
         self.mapwidget = None
 
     def hideMapping(self):
         self.dialog.mappingGroup.hide()
 
-    def showMapping(self, keepMapping=False):
+    def showMapping(self, keepMapping=False, hook=None):
         # first, check that we can read the file
         try:
             self.importer = self.importerFunc(self.parent.deck, self.file)
+            if hook:
+                hook()
             if not keepMapping:
                 self.mapping = self.importer.mapping
         except ImportFormatError, e:
@@ -171,6 +263,7 @@ class ImportDialog(QDialog):
             self.dialog.mappingArea.show()
         else:
             self.dialog.mappingArea.hide()
+            self.dialog.updateButton.hide()
             return
         # set up the mapping grid
         if self.mapwidget:
@@ -180,8 +273,8 @@ class ImportDialog(QDialog):
         self.mapbox.addWidget(self.mapwidget)
         self.grid = QGridLayout(self.mapwidget)
         self.mapwidget.setLayout(self.grid)
-        self.grid.setMargin(6)
-        self.grid.setSpacing(12)
+        self.grid.setMargin(3)
+        self.grid.setSpacing(6)
         fields = self.importer.fields()
         for num in range(len(self.mapping)):
             text = _("Field <b>%d</b> of file is:") % (num + 1)
@@ -207,4 +300,17 @@ class ImportDialog(QDialog):
         except ValueError:
             pass
         self.mapping[n] = f
-        self.showMapping(keepMapping=True)
+        if getattr(self.importer, "delimiter", False):
+            self.savedDelimiter = self.importer.delimiter
+            def updateDelim():
+                self.importer.delimiter = self.savedDelimiter
+            self.showMapping(hook=updateDelim, keepMapping=True)
+        else:
+            self.showMapping(keepMapping=True)
+
+    def reject(self):
+        self.modelChooser.deinit()
+        QDialog.reject(self)
+
+    def helpRequested(self):
+        QDesktopServices.openUrl(QUrl(ankiqt.appWiki + "FileImport"))
