@@ -5,13 +5,14 @@
 from PyQt4.QtGui import *
 from PyQt4.QtCore import *
 import anki, anki.utils
-from anki.sound import playFromText, stripSounds
-from anki.latex import renderLatex, stripLatex
+from anki.sound import playFromText
 from anki.utils import stripHTML
-from anki.hooks import runHook
-import types, time, re, os, urllib, sys
+from anki.hooks import runHook, runFilter
+from anki.media import stripMedia, escapeImages
+import types, time, re, os, urllib, sys, difflib
+import unicodedata as ucd
 from ankiqt import ui
-from ankiqt.ui.utils import mungeQA
+from ankiqt.ui.utils import mungeQA, getBase
 from anki.utils import fmtTimeSpan
 from PyQt4.QtWebKit import QWebPage, QWebView
 
@@ -52,11 +53,8 @@ class View(object):
         "Idempotently display the current state (prompt for question, etc)"
         if self.state == "noDeck" or self.state == "studyScreen":
             return
-        self.clearWindow()
-        self.haveTop = (self.main.lastCard and (
-            self.main.config['showLastCardContent'] or
-            self.main.config['showLastCardInterval'])) or (
-            self.main.currentCard and self.main.currentCard.due > time.time())
+        self.buffer = ""
+        self.haveTop = self.needFutureWarning()
         self.drawRule = (self.main.config['qaDivider'] and
                          self.main.currentCard and
                          not self.main.currentCard.cardModel.questionInAnswer)
@@ -64,10 +62,12 @@ class View(object):
             if self.haveTop:
                 self.drawTopSection()
         if self.state == "showQuestion":
+            self.setBackground()
             self.drawQuestion()
             if self.drawRule:
                 self.write("<hr>")
         elif self.state == "showAnswer":
+            self.setBackground()
             if not self.main.currentCard.cardModel.questionInAnswer:
                 self.drawQuestion(nosound=True)
             if self.drawRule:
@@ -84,7 +84,7 @@ class View(object):
         s = "<style>\n"
         if self.main.deck:
             s += self.main.deck.css
-        s += "div { white-space: pre-wrap; }"
+        s = runFilter("addStyles", s, self.main.currentCard)
         s += "</style>"
         return s
 
@@ -92,6 +92,10 @@ class View(object):
         self.body.setHtml("")
         self.buffer = ""
 
+    def setBackground(self):
+        col = self.main.currentCard.cardModel.lastFontColour
+        self.write("<style>html { background: %s;}</style>" % col)
+
     # Font properties & output
     ##########################################################################
 
@@ -100,8 +104,19 @@ class View(object):
         self.buffer = self.addStyles() + self.buffer
         # hook for user css
         runHook("preFlushHook")
+        self.buffer = '''<html><head>%s</head><body>%s</body></html>''' % (
+            getBase(self.main.deck, self.main.currentCard), self.buffer)
         #print self.buffer.encode("utf-8")
-        self.body.setHtml(self.buffer)
+        b = self.buffer
+        # Feeding webkit unicode can result in it not finding images, so on
+        # linux/osx we percent escape the image paths as utf8. On Windows the
+        # problem is more complicated - if we percent-escape as utf8 it fixes
+        # some images but breaks others. When filenames are normalized by
+        # dropbox they become unreadable if we escape them.
+        if not sys.platform.startswith("win32"):
+            # and self.main.config['mediaLocation'] == "dropbox"):
+            b = escapeImages(b)
+        self.body.setHtml(b)
 
     def write(self, text):
         if type(text) != types.UnicodeType:
@@ -126,45 +141,100 @@ class View(object):
         q = self.main.currentCard.htmlQuestion()
         if self.haveTop:
             height = 35
+        elif self.main.currentCard.cardModel.questionInAnswer:
+            height = 40
         else:
             height = 45
+        q = runFilter("drawQuestion", q, self.main.currentCard)
         self.write(self.center(self.mungeQA(self.main.deck, q), height))
-        if self.state != self.oldState and not nosound:
+        if (self.state != self.oldState and not nosound
+            and self.main.config['autoplaySounds']):
             playFromText(q)
 
+    def calculateOkBadStyle(self):
+        "Precalculates styles for correct and incorrect part of answer"
+        sz = 20
+        fn = u"Arial"
+        for fm in self.main.currentCard.fact.model.fieldModels:
+            if fm.name == self.main.currentCard.cardModel.typeAnswer:
+                sz = fm.quizFontSize or sz
+                fn = fm.quizFontFamily or fn
+                break
+        st = "background: %s; color: #000; font-size: %dpx; font-family: %s;"
+        self.styleOk  = st % (passedCharColour, sz, fn)
+        self.styleBad = st % (failedCharColour, sz, fn)
+
+    def ok(self, a):
+        "returns given sring in style correct (green)"
+        if len(a) == 0:
+            return ""
+        return "<span style='%s'>%s</span>" % (self.styleOk, a)
+
+    def bad(self, a):
+        "returns given sring in style incorrect (red)"
+        if len(a) == 0:
+            return ""
+        return "<span style='%s'>%s</span>" % (self.styleBad, a)
+
+    def head(self, a):
+        return a[:len(a) - 1]
+
+    def tail(self, a):
+        return a[len(a) - 1:]
+
+    def applyStyle(self, testChar, correct, wrong):
+        "Calculates answer fragment depending on testChar's unicode category"
+        ZERO_SIZE = 'Mn'
+        if ucd.category(testChar) == ZERO_SIZE:
+            return self.ok(self.head(correct)) + self.bad(self.tail(correct) + wrong)
+        return self.ok(correct) + self.bad(wrong)
+
+    def correct(self, a, b):
+        "Diff-corrects the typed-in answer."
+        if b == "":
+            return "";
+
+        self.calculateOkBadStyle()
+
+        ret = ""
+        lastEqual = ""
+        s = difflib.SequenceMatcher(None, b, a)
+
+        for tag, i1, i2, j1, j2 in s.get_opcodes():
+            if tag == "equal":
+                lastEqual = b[i1:i2]
+            elif tag == "replace":
+                ret += self.applyStyle(b[i1], lastEqual, 
+                                 b[i1:i2] + ("-" * ((j2 - j1) - (i2 - i1))))
+                lastEqual = ""
+            elif tag == "delete":
+                ret += self.applyStyle(b[i1], lastEqual, b[i1:i2])
+                lastEqual = ""
+            elif tag == "insert":
+                dashNum = (j2 - j1) if ucd.category(a[j1]) != 'Mn' else ((j2 - j1) - 1)
+                ret += self.applyStyle(a[j1], lastEqual, "-" * dashNum)
+                lastEqual = ""
+
+        return ret + self.ok(lastEqual)
+
     def drawAnswer(self):
         "Show the answer."
         a = self.main.currentCard.htmlAnswer()
+        a = runFilter("drawAnswer", a, self.main.currentCard)
         if self.main.currentCard.cardModel.typeAnswer:
             try:
-                cor = stripHTML(self.main.currentCard.fact[
-                    self.main.currentCard.cardModel.typeAnswer])
+                cor = stripMedia(stripHTML(self.main.currentCard.fact[
+                    self.main.currentCard.cardModel.typeAnswer]))
             except KeyError:
+                self.main.currentCard.cardModel.typeAnswer = ""
                 cor = ""
             if cor:
                 given = unicode(self.main.typeAnswerField.text())
-                res = []
-                if len(given) < len(cor):
-                    given += " " * (len(cor) - len(given))
-                sz = self.main.currentCard.cardModel.answerFontSize
-                ok = "background: %s; color: #000; font-size: %dpx" % (
-                    passedCharColour, sz)
-                bad = "background: %s; color: #000; font-size: %dpx;" % (
-                    failedCharColour, sz)
-                for (i, c) in enumerate(given):
-                    try:
-                        yes = c == cor[i]
-                    except IndexError:
-                        yes = False
-                    if yes:
-                        res.append(
-                            "<span style='%s'>%s</span>" % (ok, c))
-                    else:
-                        res.append("<span style='%s'>%s</span>" % (bad, c))
-                a = "".join(res) + "<br>" + a
-        self.write(self.center('<span id=answer />' +
-                               self.mungeQA(self.main.deck, a)))
-        if self.state != self.oldState:
+                res = self.correct(cor, given)
+                a = res + "<br>" + a
+        self.write(self.center('<span id=answer />'
+                               + self.mungeQA(self.main.deck, a)))
+        if self.state != self.oldState and self.main.config['autoplaySounds']:
             playFromText(a)
 
     def mungeQA(self, deck, txt):
@@ -174,7 +244,7 @@ class View(object):
             txt = txt.replace("</span>", "&#8203;</span>")
         return txt
 
-    def onLoadFinished(self):
+    def onLoadFinished(self, bool):
         if self.state == "showAnswer":
             if self.main.config['scrollToAnswer']:
                 mf = self.body.page().mainFrame()
@@ -187,54 +257,33 @@ class View(object):
         "Show previous card, next scheduled time, and stats."
         self.buffer += "<center>"
         self.drawFutureWarning()
-        self.drawLastCard()
         self.buffer += "</center>"
 
-    def drawFutureWarning(self):
+    def needFutureWarning(self):
         if not self.main.currentCard:
             return
-        if self.main.currentCard.due <= time.time():
+        if self.main.currentCard.due <= self.main.deck.dueCutoff:
             return
         if self.main.currentCard.due - time.time() <= self.main.deck.delay0:
             return
+        if self.main.deck.scheduler == "cram":
+            return
+        return True
+
+    def drawFutureWarning(self):
+        if not self.needFutureWarning():
+            return
         self.write("<span style='color: %s'>" % futureWarningColour +
                    _("This card was due in %s.") % fmtTimeSpan(
-            self.main.currentCard.due - time.time()) +
+            self.main.currentCard.due - time.time(), after=True) +
                    "</span>")
 
-    def drawLastCard(self):
-        "Show the last card if not the current one, and next time."
-        if self.main.lastCard:
-            if self.main.config['showLastCardContent']:
-                if (self.state == "deckFinished" or
-                    self.main.currentCard.id != self.main.lastCard.id):
-                    q = self.main.lastCard.question.replace("<br>", "  ")
-                    q = stripHTML(q)
-                    if len(q) > 50:
-                        q = q[:50] + "..."
-                    a = self.main.lastCard.answer.replace("<br>", "  ")
-                    a = stripHTML(a)
-                    if len(a) > 50:
-                        a = a[:50] + "..."
-                    s = "%s<br>%s" % (q, a)
-                    s = stripLatex(s)
-                    self.write('<span class="lastCard">%s</span><br>' % s)
-            if self.main.config['showLastCardInterval']:
-                if self.main.lastQuality > 1:
-                    msg = _("Well done! This card will appear again in "
-                            "<b>%(next)s</b>.") % \
-                            {"next":self.main.lastScheduledTime}
-                else:
-                    msg = _("This card will appear again later.")
-                self.write(msg)
-            self.write("<br>")
-
     # Welcome/empty/finished deck messages
     ##########################################################################
 
     def drawWelcomeMessage(self):
-        self.main.mainWin.welcomeText.setText(_("""
-<h1>Welcome to Anki!</h1>
+        self.main.mainWin.welcomeText.setText("""\
+<h1>%(welcome)s</h1>
 <p>
 <table>
 
@@ -242,8 +291,8 @@ class View(object):
 <td width=50>
 <a href="welcome:addfacts"><img src=":/icons/list-add.png"></a>
 </td>
-<td valign=middle><h1><a href="welcome:addfacts">Add material</a></h1>
-Start adding your own material.</td>
+<td valign=middle><h1><a href="welcome:addfacts">%(add)s</a></h1>
+%(start)s</td>
 </tr>
 
 </table>
@@ -251,35 +300,19 @@ Start adding your own material.</td>
 <br>
 <table>
 
-<tr>
-<td>
-<a href="welcome:open"><img src=":/icons/document-open.png"></a>
-</td>
-<td valign=middle><h2><a href="welcome:open">Open Local Deck</a></h2></td>
-</tr>
-
-<tr>
-<td>
-<a href="welcome:openrem"><img src=":/icons/document-open-remote.png"></a>
-</td>
-<td valign=middle><h2><a href="welcome:openrem">Open Online Deck</a></h2></td>
-</tr>
-
 <tr>
 <td width=50>
-<a href="welcome:sample"><img src=":/icons/anki.png"></a>
+<a href="welcome:back"><img src=":/icons/go-previous.png"></a>
 </td>
-<td valign=middle><h2><a href="welcome:sample">Open Sample Deck</a></h2></td>
+<td valign=middle><h2><a href="welcome:back">%(back)s</a></h2></td>
 </tr>
 
-<tr>
-<td width=50>
-<a href="welcome:more"><img src=":/icons/khtml_kget.png"></a>
-</td>
-<td valign=middle><h2><a href="welcome:more">Get More Decks</a></h2></td>
-</tr>
-
-</table>"""))
+</table>""" % \
+	{"welcome":_("Welcome to Anki!"),
+         "add":_("Add Material"),
+         "start":_("Start adding your own material."),
+         "back":_("Back to Deck Browser"),
+         })
 
     def drawDeckFinishedMessage(self):
         "Tell the user the deck is finished."
@@ -296,7 +329,7 @@ class AnkiWebView(QWebView):
         if evt.matches(QKeySequence.Copy):
             self.triggerPageAction(QWebPage.Copy)
             evt.accept()
-        evt.ignore()
+        QWebView.keyPressEvent(self, evt)
 
     def contextMenuEvent(self, evt):
         QWebView.contextMenuEvent(self, evt)
